@@ -4,18 +4,19 @@
 //
 // Copyright (c) DUSK NETWORK. All rights reserved.
 
+use std::convert::TryInto;
 use std::net::SocketAddr;
-use std::{convert::TryInto, sync::Arc};
 
 use tokio::sync::mpsc::{Receiver, Sender};
-use tokio::sync::RwLock;
 use tracing::*;
 
-use crate::encoding::message::{BroadcastPayload, Message, NodePayload};
-use crate::kbucket::Tree;
+use crate::encoding::message::{
+    BroadcastPayload, Header, Message, NodePayload,
+};
+use crate::kbucket::{BinaryKey, NodeInsertError, Tree};
 use crate::peer::{PeerInfo, PeerNode};
 use crate::transport::{MessageBeanIn, MessageBeanOut};
-use crate::K_K;
+use crate::{RwLock, K_K};
 
 /// Message metadata for incoming message notifications
 #[derive(Debug)]
@@ -39,34 +40,54 @@ pub(crate) struct MessageHandler;
 
 impl MessageHandler {
     pub(crate) fn start(
-        ktable: Arc<RwLock<Tree<PeerInfo>>>,
+        ktable: RwLock<Tree<PeerInfo>>,
         mut inbound_receiver: Receiver<MessageBeanIn>,
         outbound_sender: Sender<MessageBeanOut>,
         listener_sender: Sender<(Vec<u8>, MessageInfo)>,
         auto_propagate: bool,
+        recursive_discovery: bool,
     ) {
+        let nodes_reply_fn = match recursive_discovery {
+            true => |header: Header, target: BinaryKey| {
+                Message::FindNodes(header, target)
+            },
+            false => |header: Header, _: BinaryKey| Message::Ping(header),
+        };
         tokio::spawn(async move {
             debug!("MessageHandler started");
+            let my_header = { ktable.read().await.root().as_header() };
             while let Some((message, mut remote_node_addr)) =
                 inbound_receiver.recv().await
             {
-                debug!("Mantainer received message {:?}", message);
+                debug!("Handler received message");
+                trace!("Handler received message {:?}", message);
                 remote_node_addr.set_port(message.header().sender_port);
                 let remote_node = PeerNode::from_socket(remote_node_addr);
-                let my_header = ktable.read().await.root().as_header();
                 match ktable.write().await.insert(remote_node) {
-                    Err(_) => {
-                        error!("Unable to insert node");
-                        continue;
-                    }
+                    Err(e) => match e {
+                        NodeInsertError::Full(n) => {
+                            debug!(
+                                "Unable to insert node - FULL {}",
+                                n.value().address()
+                            )
+                        }
+                        NodeInsertError::Invalid(n) => {
+                            error!(
+                                "Unable to insert node - INVALID {}",
+                                n.value().address()
+                            );
+                            continue;
+                        }
+                    },
                     Ok(result) => {
                         debug!("Written node in ktable: {:?}", &result);
                         if let Some(pending) = result.pending_eviction() {
                             outbound_sender
-                                .try_send((
+                                .send((
                                     Message::Ping(my_header),
                                     vec![*pending.value().address()],
                                 ))
+                                .await
                                 .unwrap_or_else(|op| {
                                     error!("Unable to send PING to pending node {:?}", op)
                                 });
@@ -76,10 +97,11 @@ impl MessageHandler {
                 match message {
                     Message::Ping(_) => {
                         outbound_sender
-                            .try_send((
+                            .send((
                                 Message::Pong(my_header),
                                 vec![remote_node_addr],
                             ))
+                            .await
                             .unwrap_or_else(|op| {
                                 error!("Unable to send Pong {:?}", op)
                             });
@@ -87,7 +109,7 @@ impl MessageHandler {
                     Message::Pong(_) => {}
                     Message::FindNodes(_, target) => {
                         outbound_sender
-                            .try_send((
+                            .send((
                                 Message::Nodes(
                                     my_header,
                                     NodePayload {
@@ -101,30 +123,57 @@ impl MessageHandler {
                                 ),
                                 vec![remote_node_addr],
                             ))
+                            .await
                             .unwrap_or_else(|op| {
                                 error!("Unable to send Nodes {:?}", op)
                             });
                     }
                     Message::Nodes(_, nodes) => {
                         if !nodes.peers.is_empty() {
-                            let targets = nodes
+                            let reader = ktable.read().await;
+                            let messages = nodes
                                 .peers
                                 .iter()
                                 //filter out my ID to avoid loopback
                                 .filter(|&n| {
                                     &n.id != my_header.binary_id.as_binary()
                                 })
-                                .map(|n| n.to_socket_address())
-                                .collect();
-                            outbound_sender
-                                .try_send((Message::Ping(my_header), targets))
-                                .unwrap_or_else(|op| {
-                                    error!("Unable to send PING {:?}", op)
-                                });
+                                .filter(|&n| {
+                                    let h = my_header
+                                        .binary_id
+                                        .calculate_distance(&n.id);
+                                    match h {
+                                        None => false,
+                                        Some(h) => {
+                                            if reader.is_bucket_full(h) {
+                                                return false;
+                                            };
+                                            reader.has_peer(&n.id).is_none()
+                                        }
+                                    }
+                                })
+                                .map(|n| {
+                                    (
+                                        nodes_reply_fn(my_header, n.id),
+                                        vec![n.to_socket_address()],
+                                    )
+                                })
+                                .collect::<Vec<(Message, Vec<SocketAddr>)>>();
+                            for tosend in messages {
+                                outbound_sender.send(tosend).await.unwrap_or_else(
+                                    |op| {
+                                        error!( "Unable to send FindNodes after reply {:?}", op)
+                                    },
+                                );
+                            }
                         }
                     }
                     Message::Broadcast(_, payload) => {
-                        debug!("Received payload with height {:?}", payload);
+                        debug!(
+                            "Received payload with height {:?} and len {}",
+                            payload.height,
+                            payload.gossip_frame.len()
+                        );
 
                         // Aggregate message + metadata for lib client
                         let msg = payload.gossip_frame.clone();
@@ -134,19 +183,19 @@ impl MessageHandler {
                         };
 
                         // Notify lib client
-                        listener_sender.try_send((msg, md)).unwrap_or_else(
+                        listener_sender.send((msg, md)).await.unwrap_or_else(
                             |op| error!("Unable to notify client {:?}", op),
                         );
                         if auto_propagate && payload.height > 0 {
-                            let table_read = ktable.read().await;
                             debug!(
                                 "Extracting for height {:?}",
                                 payload.height - 1
                             );
+                            let table_read = ktable.read().await;
 
-                            table_read
+                            let messages: Vec<(Message, Vec<SocketAddr>)> = table_read
                                 .extract(Some((payload.height - 1).into()))
-                                .for_each(|(height, nodes)| {
+                                .map(|(height, nodes)| {
                                     let msg = Message::Broadcast(
                                         my_header,
                                         BroadcastPayload {
@@ -159,15 +208,21 @@ impl MessageHandler {
                                     let targets: Vec<SocketAddr> = nodes
                                         .map(|node| *node.value().address())
                                         .collect();
-                                    outbound_sender
-                                        .try_send((msg, targets))
-                                        .unwrap_or_else(|op| {
-                                            error!(
-                                                "Unable to send broadcast {:?}",
-                                                op
-                                            )
-                                        });
-                                });
+                                    (msg, targets)
+                                }).collect();
+                            drop(table_read);
+
+                            for tosend in messages {
+                                outbound_sender
+                                    .send(tosend)
+                                    .await
+                                    .unwrap_or_else(|op| {
+                                        error!(
+                                            "Unable to send broadcast {:?}",
+                                            op
+                                        )
+                                    });
+                            }
                         }
                     }
                 }
