@@ -4,8 +4,10 @@
 //
 // Copyright (c) DUSK NETWORK. All rights reserved.
 
-use std::{convert::TryInto, net::SocketAddr, sync::Arc, time::Duration};
+use std::collections::HashMap;
+use std::{convert::TryInto, net::SocketAddr, time::Duration};
 
+use encoding::message::Header;
 use encoding::{message::Message, payload::BroadcastPayload};
 use handling::MessageHandler;
 pub use handling::MessageInfo;
@@ -13,10 +15,10 @@ use itertools::Itertools;
 use kbucket::{Tree, TreeBuilder};
 use mantainer::TableMantainer;
 use peer::{PeerInfo, PeerNode};
+pub(crate) use rwlock::RwLock;
 use tokio::sync::mpsc::{self, Receiver, Sender};
-use tokio::sync::RwLock;
 use tokio::task;
-use tracing::info;
+use tracing::{error, info};
 use transport::{MessageBeanOut, WireNetwork};
 
 mod encoding;
@@ -24,15 +26,27 @@ mod handling;
 mod kbucket;
 mod mantainer;
 mod peer;
-mod transport;
+mod rwlock;
+pub mod transport;
 
 // Max amount of nodes a bucket should contain
-const K_K: usize = 20;
+const DEFAULT_K_K: usize = 20;
+const K_K: usize = get_k_k();
 const K_ID_LEN_BYTES: usize = 16;
 const K_NONCE_LEN: usize = 4;
 const K_DIFF_MIN_BIT: usize = 8;
 const K_DIFF_PRODUCED_BIT: usize = 8;
-const MAX_DATAGRAM_SIZE: usize = 65_507;
+const DEFAULT_CHANNEL_SIZE: usize = 100;
+
+const fn get_k_k() -> usize {
+    match option_env!("KADCAST_K") {
+        Some(v) => match konst::primitive::parse_usize(v) {
+            Ok(e) => e,
+            Err(_) => DEFAULT_K_K,
+        },
+        None => DEFAULT_K_K,
+    }
+}
 
 // Redundacy factor for lookup
 const K_ALPHA: usize = 3;
@@ -55,7 +69,8 @@ pub const ENABLE_BROADCAST_PROPAGATION: bool = true;
 /// Struct representing the Kadcast Network Peer
 pub struct Peer {
     outbound_sender: Sender<MessageBeanOut>,
-    ktable: Arc<RwLock<Tree<PeerInfo>>>,
+    ktable: RwLock<Tree<PeerInfo>>,
+    header: Header,
 }
 
 /// [NetworkListen] is notified each time a broadcasted
@@ -69,17 +84,25 @@ impl Peer {
         listen_address: String,
         bootstrapping_nodes: Vec<String>,
         auto_propagate: bool,
+        recursive_discovery: bool,
+        channel_size: usize,
         listener: L,
         tree: Tree<PeerInfo>,
+        transport_conf: HashMap<String, String>,
     ) -> Self {
-        let (inbound_channel_tx, inbound_channel_rx) = mpsc::channel(32);
-        let (outbound_channel_tx, outbound_channel_rx) = mpsc::channel(32);
-        let (notification_channel_tx, listener_channel_rx) = mpsc::channel(32);
+        let (inbound_channel_tx, inbound_channel_rx) =
+            mpsc::channel(channel_size);
+        let (outbound_channel_tx, outbound_channel_rx) =
+            mpsc::channel(channel_size);
+        let (notification_channel_tx, listener_channel_rx) =
+            mpsc::channel(channel_size);
 
-        let table = Arc::new(RwLock::new(tree));
+        let header = tree.root().as_header();
+        let table = RwLock::new(tree, Duration::from_secs(1));
         let peer = Peer {
             outbound_sender: outbound_channel_tx.clone(),
             ktable: table.clone(),
+            header,
         };
         MessageHandler::start(
             table.clone(),
@@ -87,12 +110,15 @@ impl Peer {
             outbound_channel_tx.clone(),
             notification_channel_tx,
             auto_propagate,
+            recursive_discovery,
         );
-        tokio::spawn(WireNetwork::start(
+        WireNetwork::start(
             inbound_channel_tx,
             listen_address,
             outbound_channel_rx,
-        ));
+            transport_conf,
+            channel_size,
+        );
         tokio::spawn(TableMantainer::start(
             bootstrapping_nodes,
             table,
@@ -145,19 +171,33 @@ impl Peer {
     /// system. It **does not guarantee** the message will be broadcasted
     pub async fn broadcast(&self, message: &[u8], height: Option<usize>) {
         if message.is_empty() {
+            error!("Message empty");
             return;
         }
-        for (h, nodes) in self.ktable.read().await.extract(height) {
-            let msg = Message::Broadcast(
-                self.ktable.read().await.root().as_header(),
-                BroadcastPayload {
-                    height: h.try_into().unwrap(),
-                    gossip_frame: message.to_vec(), //FIX_ME: avoid clone
-                },
-            );
-            let targets: Vec<SocketAddr> =
-                nodes.map(|node| *node.value().address()).collect();
-            let _ = self.outbound_sender.send((msg, targets)).await;
+
+        let tosend: Vec<(Message, Vec<SocketAddr>)> = self
+            .ktable
+            .read()
+            .await
+            .extract(height)
+            .map(|(h, nodes)| {
+                let msg = Message::Broadcast(
+                    self.header,
+                    BroadcastPayload {
+                        height: h.try_into().unwrap(),
+                        gossip_frame: message.to_vec(), //FIX_ME: avoid clone
+                    },
+                );
+                let targets: Vec<SocketAddr> =
+                    nodes.map(|node| *node.value().address()).collect();
+                (msg, targets)
+            })
+            .collect();
+
+        for i in tosend {
+            self.outbound_sender.send(i).await.unwrap_or_else(|e| {
+                error!("Unable to send from broadcast {}", e)
+            });
         }
     }
 
@@ -178,14 +218,19 @@ impl Peer {
         // We use the Broadcast message type while setting height to 0
         // to prevent further propagation at the receiver
         let msg = Message::Broadcast(
-            self.ktable.read().await.root().as_header(),
+            self.header,
             BroadcastPayload {
                 height: 0,
                 gossip_frame: message.to_vec(), //FIX_ME: avoid clone
             },
         );
         let targets = vec![target];
-        let _ = self.outbound_sender.send((msg, targets)).await;
+        self.outbound_sender
+            .send((msg, targets))
+            .await
+            .unwrap_or_else(|e| {
+                error!("Unable to send from send method {}", e)
+            });
     }
 
     /// Instantiate a [PeerBuilder].
@@ -216,6 +261,9 @@ pub struct PeerBuilder<L: NetworkListen + 'static> {
     listen_address: Option<String>,
     bootstrapping_nodes: Vec<String>,
     listener: L,
+    transport_conf: HashMap<String, String>,
+    channel_size: usize,
+    recursive_discovery: bool,
 }
 
 impl<L: NetworkListen + 'static> PeerBuilder<L> {
@@ -259,6 +307,10 @@ impl<L: NetworkListen + 'static> PeerBuilder<L> {
         self
     }
 
+    pub fn transport_conf(&mut self) -> &mut HashMap<String, String> {
+        &mut self.transport_conf
+    }
+
     /// Enable automatic propagation of incoming broadcast messages
     ///
     /// Default value [ENABLE_BROADCAST_PROPAGATION]
@@ -270,6 +322,21 @@ impl<L: NetworkListen + 'static> PeerBuilder<L> {
         self
     }
 
+    /// Default value [DEFAULT_CHANNEL_SIZE]
+    pub fn with_channel_size(mut self, channel_size: usize) -> PeerBuilder<L> {
+        self.channel_size = channel_size;
+        self
+    }
+
+    /// Default value true
+    pub fn with_recursive_discovery(
+        mut self,
+        recursive_discovery: bool,
+    ) -> PeerBuilder<L> {
+        self.recursive_discovery = recursive_discovery;
+        self
+    }
+
     fn new(
         public_address: String,
         bootstrapping_nodes: Vec<String>,
@@ -286,6 +353,9 @@ impl<L: NetworkListen + 'static> PeerBuilder<L> {
             node_ttl: Duration::from_millis(BUCKET_DEFAULT_NODE_TTL_MILLIS),
             bucket_ttl: Duration::from_secs(BUCKET_DEFAULT_TTL_SECS),
             auto_propagate: ENABLE_BROADCAST_PROPAGATION,
+            transport_conf: transport::default_configuration(),
+            channel_size: DEFAULT_CHANNEL_SIZE,
+            recursive_discovery: true,
         }
     }
 
@@ -301,8 +371,11 @@ impl<L: NetworkListen + 'static> PeerBuilder<L> {
             self.listen_address.unwrap_or(self.public_address),
             self.bootstrapping_nodes,
             self.auto_propagate,
+            self.recursive_discovery,
+            self.channel_size,
             self.listener,
             tree,
+            self.transport_conf,
         )
     }
 }
